@@ -29,7 +29,7 @@
 #include "apc_sma.h"
 #include "apc.h"
 #include "apc_globals.h"
-#include "apc_lock.h"
+#include "apc_mutex.h"
 #include "apc_shm.h"
 #include "apc_cache.h"
 
@@ -49,7 +49,7 @@ enum {
 
 typedef struct sma_header_t sma_header_t;
 struct sma_header_t {
-	apc_lock_t sma_lock;    /* segment lock */
+	apc_mutex_t sma_lock;    /* segment lock */
 	size_t segsize;         /* size of entire segment */
 	size_t avail;           /* bytes available (not necessarily contiguous) */
 };
@@ -59,6 +59,11 @@ struct sma_header_t {
 #define SMA_RO(sma, i)   ((char*)(sma->segs[i]).roaddr)
 #define SMA_LCK(sma, i)  ((SMA_HDR(sma, i))->sma_lock)
 
+#define SMA_CREATE_LOCK  APC_CREATE_MUTEX
+#define SMA_DESTROY_LOCK APC_DESTROY_MUTEX
+#define SMA_LOCK(sma, i) APC_MUTEX_LOCK(&SMA_LCK(sma, i))
+#define SMA_UNLOCK(sma, i) APC_MUTEX_UNLOCK(&SMA_LCK(sma, i))
+
 #if 0
 /* global counter for identifying blocks
  * Technically it is possible to do the same
@@ -266,7 +271,7 @@ static APC_HOTSPOT size_t sma_deallocate(void* shmaddr, size_t offset)
 /* }}} */
 
 /* {{{ APC SMA API */
-PHP_APCU_API void apc_sma_api_init(apc_sma_t* sma, void** data, apc_sma_expunge_f expunge, int32_t num, zend_ulong size, char *mask) {
+PHP_APCU_API void apc_sma_init(apc_sma_t* sma, void** data, apc_sma_expunge_f expunge, int32_t num, zend_ulong size, char *mask) {
 	uint i;
 
 	if (sma->initialized) {
@@ -295,7 +300,7 @@ PHP_APCU_API void apc_sma_api_init(apc_sma_t* sma, void** data, apc_sma_expunge_
 
 	sma->size = size > 0 ? size : DEFAULT_SEGSIZE;
 
-	sma->segs = (apc_segment_t*) apc_emalloc((sma->num * sizeof(apc_segment_t)));
+	sma->segs = (apc_segment_t*) pemalloc(sma->num * sizeof(apc_segment_t), 1);
 
 	for (i = 0; i < sma->num; i++) {
 		sma_header_t*   header;
@@ -322,7 +327,7 @@ PHP_APCU_API void apc_sma_api_init(apc_sma_t* sma, void** data, apc_sma_expunge_
 		shmaddr = sma->segs[i].shmaddr;
 
 		header = (sma_header_t*) shmaddr;
-		CREATE_LOCK(&header->sma_lock);
+		SMA_CREATE_LOCK(&header->sma_lock);
 		header->segsize = sma->size;
 		header->avail = sma->size - ALIGNWORD(sizeof(sma_header_t)) - ALIGNWORD(sizeof(block_t)) - ALIGNWORD(sizeof(block_t));
 
@@ -356,80 +361,91 @@ PHP_APCU_API void apc_sma_api_init(apc_sma_t* sma, void** data, apc_sma_expunge_
 	}
 }
 
-PHP_APCU_API void apc_sma_api_cleanup(apc_sma_t* sma) {
+PHP_APCU_API void apc_sma_detach(apc_sma_t* sma) {
+	/* Important: This function should not clean up anything that's in shared memory,
+	 * only detach our process-local use of it. In particular locks cannot be destroyed
+	 * here. */
 	uint i;
 
 	assert(sma->initialized);
+	sma->initialized = 0;
 
 	for (i = 0; i < sma->num; i++) {
-		DESTROY_LOCK(&SMA_LCK(sma, i));
 #if APC_MMAP
 		apc_unmap(&sma->segs[i]);
 #else
 		apc_shm_detach(&sma->segs[i]);
 #endif
 	}
-	sma->initialized = 0;
 
-	apc_efree(sma->segs);
+	free(sma->segs);
 }
 
-PHP_APCU_API void* apc_sma_api_malloc_ex(apc_sma_t* sma, zend_ulong n, zend_ulong fragment, zend_ulong* allocated) {
+PHP_APCU_API void* apc_sma_malloc_ex(apc_sma_t* sma, zend_ulong n, zend_ulong fragment, zend_ulong* allocated) {
 	size_t off;
 	uint i;
-	int nuked = 0;
+	zend_bool nuked = 0;
+	int32_t last = sma->last;
 
 restart:
 	assert(sma->initialized);
 
-	WLOCK(&SMA_LCK(sma, sma->last));
+	if (!SMA_LOCK(sma, last)) {
+		return NULL;
+	}
 
-	off = sma_allocate(SMA_HDR(sma, sma->last), n, fragment, allocated);
+	off = sma_allocate(SMA_HDR(sma, last), n, fragment, allocated);
 
-	if(off == -1) {
+	if (off == -1) {
 		/* retry failed allocation after we expunge */
-		WUNLOCK(&SMA_LCK(sma, sma->last));
-		sma->expunge(
-			*(sma->data), (n+fragment));
-		WLOCK(&SMA_LCK(sma, sma->last));
-		off = sma_allocate(SMA_HDR(sma, sma->last), n, fragment, allocated);
+		SMA_UNLOCK(sma, last);
+		sma->expunge(*(sma->data), n+fragment);
+		if (!SMA_LOCK(sma, last)) {
+			return NULL;
+		}
+		off = sma_allocate(SMA_HDR(sma, last), n, fragment, allocated);
 	}
 
 	if (off != -1) {
-		void* p = (void *)(SMA_ADDR(sma, sma->last) + off);
-		WUNLOCK(&SMA_LCK(sma, sma->last));
+		void* p = (void *)(SMA_ADDR(sma, last) + off);
+		SMA_UNLOCK(sma, last);
 #ifdef VALGRIND_MALLOCLIKE_BLOCK
 		VALGRIND_MALLOCLIKE_BLOCK(p, n, 0, 0);
 #endif
 		return p;
 	}
 
-	WUNLOCK(&SMA_LCK(sma, sma->last));
+	SMA_UNLOCK(sma, last);
 
 	for (i = 0; i < sma->num; i++) {
-		if (i == sma->last) {
+		if (i == last) {
 			continue;
 		}
-		WLOCK(&SMA_LCK(sma, i));
+
+		if (!SMA_LOCK(sma, i)) {
+			return NULL;
+		}
+
 		off = sma_allocate(SMA_HDR(sma, i), n, fragment, allocated);
-		if(off == -1) {
+		if (off == -1) {
 			/* retry failed allocation after we expunge */
-			WUNLOCK(&SMA_LCK(sma, i));
-			sma->expunge(
-				*(sma->data), (n+fragment));
-			WLOCK(&SMA_LCK(sma, i));
+			SMA_UNLOCK(sma, i);
+			sma->expunge(*(sma->data), n+fragment);
+			if (!SMA_LOCK(sma, i)) {
+				return NULL;
+			}
 			off = sma_allocate(SMA_HDR(sma, i), n, fragment, allocated);
 		}
 		if (off != -1) {
 			void* p = (void *)(SMA_ADDR(sma, i) + off);
-			WUNLOCK(&SMA_LCK(sma, i));
 			sma->last = i;
+			SMA_UNLOCK(sma, i);
 #ifdef VALGRIND_MALLOCLIKE_BLOCK
 			VALGRIND_MALLOCLIKE_BLOCK(p, n, 0, 0);
 #endif
 			return p;
 		}
-		WUNLOCK(&SMA_LCK(sma, i));
+		SMA_UNLOCK(sma, i);
 	}
 
 	/* I've tried being nice, but now you're just asking for it */
@@ -444,40 +460,13 @@ restart:
 	return NULL;
 }
 
-PHP_APCU_API void* apc_sma_api_malloc(apc_sma_t* sma, zend_ulong n)
+PHP_APCU_API void* apc_sma_malloc(apc_sma_t* sma, zend_ulong n)
 {
 	zend_ulong allocated;
-	return apc_sma_api_malloc_ex(
-		sma, n, MINBLOCKSIZE, &allocated);
+	return apc_sma_malloc_ex(sma, n, MINBLOCKSIZE, &allocated);
 }
 
-PHP_APCU_API void* apc_sma_api_realloc(apc_sma_t* sma, void* p, zend_ulong n) {
-	apc_sma_api_free(sma, p);
-	return apc_sma_api_malloc(
-		sma, n);
-}
-
-PHP_APCU_API char* apc_sma_api_strdup(apc_sma_t* sma, const char* s) {
-	void* q;
-	int len;
-
-	if(!s) {
-		return NULL;
-	}
-
-	len = strlen(s)+1;
-	q = apc_sma_api_malloc(
-		sma, len);
-
-	if(!q) {
-		return NULL;
-	}
-
-	memcpy(q, s, len);
-	return q;
-}
-
-PHP_APCU_API void apc_sma_api_free(apc_sma_t* sma, void* p) {
+PHP_APCU_API void apc_sma_free(apc_sma_t* sma, void* p) {
 	uint i;
 	size_t offset;
 
@@ -490,9 +479,12 @@ PHP_APCU_API void apc_sma_api_free(apc_sma_t* sma, void* p) {
 	for (i = 0; i < sma->num; i++) {
 		offset = (size_t)((char *)p - SMA_ADDR(sma, i));
 		if (p >= (void*)SMA_ADDR(sma, i) && offset < sma->size) {
-			WLOCK(&SMA_LCK(sma, i));
+			if (!SMA_LOCK(sma, i)) {
+				return;
+			}
+
 			sma_deallocate(SMA_HDR(sma, i), offset);
-			WUNLOCK(&SMA_LCK(sma, i));
+			SMA_UNLOCK(sma, i);
 #ifdef VALGRIND_FREELIKE_BLOCK
 			VALGRIND_FREELIKE_BLOCK(p, 0);
 #endif
@@ -504,7 +496,7 @@ PHP_APCU_API void apc_sma_api_free(apc_sma_t* sma, void* p) {
 }
 
 #ifdef APC_MEMPROTECT
-PHP_APCU_API void* apc_sma_api_protect(apc_sma_t* sma, void* p) {
+PHP_APCU_API void* apc_sma_protect(apc_sma_t* sma, void* p) {
 	unsigned int i = 0;
 	size_t offset;
 
@@ -530,7 +522,7 @@ PHP_APCU_API void* apc_sma_api_protect(apc_sma_t* sma, void* p) {
 	return NULL;
 }
 
-PHP_APCU_API void* apc_sma_api_unprotect(apc_sma_t* sma, void* p){
+PHP_APCU_API void* apc_sma_unprotect(apc_sma_t* sma, void* p){
 	unsigned int i = 0;
 	size_t offset;
 
@@ -556,11 +548,11 @@ PHP_APCU_API void* apc_sma_api_unprotect(apc_sma_t* sma, void* p){
 	return NULL;
 }
 #else
-PHP_APCU_API void* apc_sma_api_protect(apc_sma_t* sma, void *p) { return p; }
-PHP_APCU_API void* apc_sma_api_unprotect(apc_sma_t* sma, void *p) { return p; }
+PHP_APCU_API void* apc_sma_protect(apc_sma_t* sma, void *p) { return p; }
+PHP_APCU_API void* apc_sma_unprotect(apc_sma_t* sma, void *p) { return p; }
 #endif
 
-PHP_APCU_API apc_sma_info_t* apc_sma_api_info(apc_sma_t* sma, zend_bool limited) {
+PHP_APCU_API apc_sma_info_t* apc_sma_info(apc_sma_t* sma, zend_bool limited) {
 	apc_sma_info_t* info;
 	apc_sma_link_t** link;
 	uint i;
@@ -571,11 +563,11 @@ PHP_APCU_API apc_sma_info_t* apc_sma_api_info(apc_sma_t* sma, zend_bool limited)
 		return NULL;
 	}
 
-	info = (apc_sma_info_t*) apc_emalloc(sizeof(apc_sma_info_t));
+	info = (apc_sma_info_t*) pemalloc(sizeof(apc_sma_info_t), 1);
 	info->num_seg = sma->num;
 	info->seg_size = sma->size - (ALIGNWORD(sizeof(sma_header_t)) + ALIGNWORD(sizeof(block_t)) + ALIGNWORD(sizeof(block_t)));
 
-	info->list = apc_emalloc(info->num_seg * sizeof(apc_sma_link_t*));
+	info->list = pemalloc(info->num_seg * sizeof(apc_sma_link_t*), 1);
 	for (i = 0; i < sma->num; i++) {
 		info->list[i] = NULL;
 	}
@@ -586,7 +578,7 @@ PHP_APCU_API apc_sma_info_t* apc_sma_api_info(apc_sma_t* sma, zend_bool limited)
 
 	/* For each segment */
 	for (i = 0; i < sma->num; i++) {
-		RLOCK(&SMA_LCK(sma, i));
+		SMA_LOCK(sma, i);
 		shmaddr = SMA_ADDR(sma, i);
 		prv = BLOCKAT(ALIGNWORD(sizeof(sma_header_t)));
 
@@ -598,7 +590,7 @@ PHP_APCU_API apc_sma_info_t* apc_sma_api_info(apc_sma_t* sma, zend_bool limited)
 
 			CHECK_CANARY(cur);
 
-			*link = apc_emalloc(sizeof(apc_sma_link_t));
+			*link = pemalloc(sizeof(apc_sma_link_t), 1);
 			(*link)->size = cur->size;
 			(*link)->offset = prv->fnext;
 			(*link)->next = NULL;
@@ -606,13 +598,13 @@ PHP_APCU_API apc_sma_info_t* apc_sma_api_info(apc_sma_t* sma, zend_bool limited)
 
 			prv = cur;
 		}
-		RUNLOCK(&SMA_LCK(sma, i));
+		SMA_UNLOCK(sma, i);
 	}
 
 	return info;
 }
 
-PHP_APCU_API void apc_sma_api_free_info(apc_sma_t* sma, apc_sma_info_t* info) {
+PHP_APCU_API void apc_sma_free_info(apc_sma_t* sma, apc_sma_info_t* info) {
 	int i;
 
 	for (i = 0; i < info->num_seg; i++) {
@@ -620,14 +612,14 @@ PHP_APCU_API void apc_sma_api_free_info(apc_sma_t* sma, apc_sma_info_t* info) {
 		while (p) {
 			apc_sma_link_t* q = p;
 			p = p->next;
-			apc_efree(q);
+			free(q);
 		}
 	}
-	apc_efree(info->list);
-	apc_efree(info);
+	free(info->list);
+	free(info);
 }
 
-PHP_APCU_API zend_ulong apc_sma_api_get_avail_mem(apc_sma_t* sma) {
+PHP_APCU_API zend_ulong apc_sma_get_avail_mem(apc_sma_t* sma) {
 	size_t avail_mem = 0;
 	uint i;
 
@@ -638,7 +630,7 @@ PHP_APCU_API zend_ulong apc_sma_api_get_avail_mem(apc_sma_t* sma) {
 	return avail_mem;
 }
 
-PHP_APCU_API zend_bool apc_sma_api_get_avail_size(apc_sma_t* sma, size_t size) {
+PHP_APCU_API zend_bool apc_sma_get_avail_size(apc_sma_t* sma, size_t size) {
 	uint i;
 
 	for (i = 0; i < sma->num; i++) {
@@ -650,17 +642,13 @@ PHP_APCU_API zend_bool apc_sma_api_get_avail_size(apc_sma_t* sma, size_t size) {
 	return 0;
 }
 
-PHP_APCU_API void apc_sma_api_check_integrity(apc_sma_t* sma)
+PHP_APCU_API void apc_sma_check_integrity(apc_sma_t* sma)
 {
 	/* dummy */
 }
 
-/* {{{ APC SMA */
-apc_sma_api_impl(apc_sma, &apc_user_cache, apc_cache_default_expunge);
 /* }}} */
 
- /* }}} */
-
 /*
  * Local variables:
  * tab-width: 4
