@@ -34,6 +34,8 @@
 #include "apc_iterator.h"
 #include "apc_sma.h"
 #include "apc_lock.h"
+#include "apc_mutex.h"
+#include "apc_strings.h"
 #include "php_globals.h"
 #include "php_ini.h"
 #include "ext/standard/info.h"
@@ -76,13 +78,17 @@ ZEND_DECLARE_MODULE_GLOBALS(apcu)
 apc_cache_t* apc_user_cache = NULL;
 
 /* External APC SMA */
-apc_sma_api_extern(apc_sma);
+apc_sma_t apc_sma;
+
+#define X(str) zend_string *apc_str_ ## str;
+	APC_STRINGS
+#undef X
 
 /* Global init functions */
 static void php_apc_init_globals(zend_apcu_globals* apcu_globals)
 {
 	apcu_globals->initialized = 0;
-	apcu_globals->slam_defense = 1;
+	apcu_globals->slam_defense = 0;
 	apcu_globals->smart = 0;
 	apcu_globals->preload_path = NULL;
 	apcu_globals->coredump_unmap = 0;
@@ -141,12 +147,11 @@ STD_PHP_INI_ENTRY("apc.smart",          "0",    PHP_INI_SYSTEM, OnUpdateLong,
 STD_PHP_INI_ENTRY("apc.mmap_file_mask",  NULL,  PHP_INI_SYSTEM, OnUpdateString,            mmap_file_mask,   zend_apcu_globals, apcu_globals)
 #endif
 STD_PHP_INI_BOOLEAN("apc.enable_cli",   "0",    PHP_INI_SYSTEM, OnUpdateBool,              enable_cli,       zend_apcu_globals, apcu_globals)
-STD_PHP_INI_BOOLEAN("apc.slam_defense", "1",    PHP_INI_SYSTEM, OnUpdateBool,              slam_defense,     zend_apcu_globals, apcu_globals)
+STD_PHP_INI_BOOLEAN("apc.slam_defense", "0",    PHP_INI_SYSTEM, OnUpdateBool,              slam_defense,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.preload_path", (char*)NULL,              PHP_INI_SYSTEM, OnUpdateString,       preload_path,  zend_apcu_globals, apcu_globals)
 STD_PHP_INI_BOOLEAN("apc.coredump_unmap", "0", PHP_INI_SYSTEM, OnUpdateBool, coredump_unmap, zend_apcu_globals, apcu_globals)
 STD_PHP_INI_BOOLEAN("apc.use_request_time", "1", PHP_INI_ALL, OnUpdateBool, use_request_time,  zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.serializer", "php", PHP_INI_SYSTEM, OnUpdateStringUnempty, serializer_name, zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.writable", "/tmp", PHP_INI_SYSTEM, OnUpdateStringUnempty, writable, zend_apcu_globals, apcu_globals)
+STD_PHP_INI_ENTRY("apc.serializer", "default", PHP_INI_SYSTEM, OnUpdateStringUnempty, serializer_name, zend_apcu_globals, apcu_globals)
 PHP_INI_END()
 
 /* }}} */
@@ -215,8 +220,15 @@ static PHP_MINIT_FUNCTION(apcu)
 
 	REGISTER_INI_ENTRIES();
 
+#define X(str) \
+	apc_str_ ## str = zend_new_interned_string( \
+		zend_string_init(#str, sizeof(#str) - 1, 1));
+	APC_STRINGS
+#undef X
+
 	/* locks initialized regardless of settings */
 	apc_lock_init();
+	APC_MUTEX_INIT();
 
 	/* Disable APC in cli mode unless overridden by apc.enable_cli */
 	if (!APCG(enable_cli) && !strcmp(sapi_module.name, "cli")) {
@@ -227,15 +239,19 @@ static PHP_MINIT_FUNCTION(apcu)
 	if (APCG(enabled)) {
 
 		if (!APCG(initialized)) {
+#if APC_MMAP
+			char *mmap_file_mask = APCG(mmap_file_mask);
+#else
+			char *mmap_file_mask = NULL;
+#endif
+
 			/* ensure this runs only once */
 			APCG(initialized) = 1;
 
 			/* initialize shared memory allocator */
-#if APC_MMAP
-			apc_sma.init(APCG(shm_segments), APCG(shm_size), APCG(mmap_file_mask));
-#else
-			apc_sma.init(APCG(shm_segments), APCG(shm_size), NULL);
-#endif
+			apc_sma_init(
+				&apc_sma, (void **) &apc_user_cache, (apc_sma_expunge_f) apc_cache_default_expunge,
+				APCG(shm_segments), APCG(shm_size), mmap_file_mask);
 
 			REGISTER_LONG_CONSTANT(APC_SERIALIZER_CONSTANT, (zend_long)&_apc_register_serializer, CONST_PERSISTENT | CONST_CS);
 
@@ -252,9 +268,6 @@ static PHP_MINIT_FUNCTION(apcu)
 				apc_find_serializer(APCG(serializer_name)),
 				APCG(entries_hint), APCG(gc_ttl), APCG(ttl), APCG(smart), APCG(slam_defense));
 
-			/* initialize pooling */
-			apc_pool_init();
-
 			/* preload data from path specified in configuration */
 			if (APCG(preload_path)) {
 				apc_cache_preload(
@@ -273,17 +286,20 @@ static PHP_MINIT_FUNCTION(apcu)
 /* {{{ PHP_MSHUTDOWN_FUNCTION(apcu) */
 static PHP_MSHUTDOWN_FUNCTION(apcu)
 {
+#define X(str) zend_string_release(apc_str_ ## str);
+	APC_STRINGS
+#undef X
+
 	/* locks shutdown regardless of settings */
 	apc_lock_cleanup();
+	APC_MUTEX_CLEANUP();
 
 	/* only shut down if APC is enabled */
 	if (APCG(enabled)) {
 		if (APCG(initialized)) {
-
-			/* destroy cache pointer */
-			apc_cache_destroy(apc_user_cache);
-			/* cleanup shared memory */
-			apc_sma.cleanup();
+			/* Detach cache and shared memory allocator from shared memory. */
+			apc_cache_detach(apc_user_cache);
+			apc_sma_detach(&apc_sma);
 
 			APCG(initialized) = 0;
 		}
@@ -293,6 +309,8 @@ static PHP_MSHUTDOWN_FUNCTION(apcu)
 #endif
 	}
 
+	apc_iterator_shutdown(module_number);
+
 	UNREGISTER_INI_ENTRIES();
 	return SUCCESS;
 } /* }}} */
@@ -304,6 +322,7 @@ static PHP_RINIT_FUNCTION(apcu)
 	ZEND_TSRMLS_CACHE_UPDATE();
 #endif
 
+	APCG(request_time) = 0;
 	if (APCG(enabled)) {
 		if (APCG(serializer_name)) {
 			/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
@@ -318,7 +337,7 @@ static PHP_RINIT_FUNCTION(apcu)
 }
 /* }}} */
 
-/* {{{ proto void apc_clear_cache() */
+/* {{{ proto void apcu_clear_cache() */
 PHP_FUNCTION(apcu_clear_cache)
 {
 	if (zend_parse_parameters_none() == FAILURE) {
@@ -330,7 +349,7 @@ PHP_FUNCTION(apcu_clear_cache)
 }
 /* }}} */
 
-/* {{{ proto array apc_cache_info([bool limited]) */
+/* {{{ proto array apcu_cache_info([bool limited]) */
 PHP_FUNCTION(apcu_cache_info)
 {
 	zend_bool limited = 0;
@@ -346,7 +365,7 @@ PHP_FUNCTION(apcu_cache_info)
 }
 /* }}} */
 
-/* {{{ */
+/* {{{ proto array apcu_key_info(string key) */
 PHP_FUNCTION(apcu_key_info)
 {
 	zend_string *key;
@@ -358,7 +377,7 @@ PHP_FUNCTION(apcu_key_info)
 	apc_cache_stat(apc_user_cache, key, return_value);
 } /* }}} */
 
-/* {{{ proto array apc_sma_info([bool limited]) */
+/* {{{ proto array apcu_sma_info([bool limited]) */
 PHP_FUNCTION(apcu_sma_info)
 {
 	apc_sma_info_t* info;
@@ -370,7 +389,7 @@ PHP_FUNCTION(apcu_sma_info)
 		return;
 	}
 
-	info = apc_sma.info(limited);
+	info = apc_sma_info(&apc_sma, limited);
 
 	if (!info) {
 		php_error_docref(NULL, E_WARNING, "No APC SMA info available.  Perhaps APC is disabled via apc.enabled?");
@@ -380,10 +399,10 @@ PHP_FUNCTION(apcu_sma_info)
 
 	add_assoc_long(return_value, "num_seg", info->num_seg);
 	add_assoc_double(return_value, "seg_size", (double)info->seg_size);
-	add_assoc_double(return_value, "avail_mem", (double)apc_sma.get_avail_mem());
+	add_assoc_double(return_value, "avail_mem", (double)apc_sma_get_avail_mem(&apc_sma));
 
 	if (limited) {
-		apc_sma.free_info(info);
+		apc_sma_free_info(&apc_sma, info);
 		return;
 	}
 
@@ -406,27 +425,21 @@ PHP_FUNCTION(apcu_sma_info)
 		add_next_index_zval(&block_lists, &list);
 	}
 	add_assoc_zval(return_value, "block_lists", &block_lists);
-	apc_sma.free_info(info);
+	apc_sma_free_info(&apc_sma, info);
 }
 /* }}} */
 
 /* {{{ php_apc_update  */
-int php_apc_update(zend_string *key, apc_cache_updater_t updater, void* data)
+zend_bool php_apc_update(
+		zend_string *key, apc_cache_updater_t updater, void *data,
+		zend_bool insert_if_not_found, time_t ttl)
 {
-	if (!APCG(enabled)) {
-		return 0;
-	}
-
 	if (APCG(serializer_name)) {
 		/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
 		apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 	}
 
-	if (!apc_cache_update(apc_user_cache, key, updater, data)) {
-		return 0;
-	}
-
-	return 1;
+	return apc_cache_update(apc_user_cache, key, updater, data, insert_if_not_found, ttl);
 }
 /* }}} */
 
@@ -442,10 +455,6 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 		return;
 	}
 
-	if (!APCG(enabled)) {
-		RETURN_FALSE;
-	}
-
 	if (APCG(serializer_name)) {
 		/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
 		apc_cache_serializer(apc_user_cache, APCG(serializer_name));
@@ -457,15 +466,18 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 		zend_ulong hkey_idx;
 		HashTable* hash = Z_ARRVAL_P(key);
 
-		/* note: only indicative of error */
+		/* We only insert keys that failed */
+		zval fail_zv;
+		ZVAL_LONG(&fail_zv, -1);
 		array_init(return_value);
+
 		ZEND_HASH_FOREACH_KEY_VAL(hash, hkey_idx, hkey, hentry) {
 			if (hkey) {
 				if (!apc_cache_store(apc_user_cache, hkey, hentry, (uint32_t) ttl, exclusive)) {
-					add_assoc_long_ex(return_value, hkey->val, hkey->len, -1);  /* -1: insertion error */
+					zend_hash_add_new(Z_ARRVAL_P(return_value), hkey, &fail_zv);
 				}
 			} else {
-				add_index_long(return_value, hkey_idx, -1);  /* -1: insertion error */
+				zend_hash_index_add_new(Z_ARRVAL_P(return_value), hkey_idx, &fail_zv);
 			}
 		} ZEND_HASH_FOREACH_END();
 		return;
@@ -474,16 +486,12 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 			/* nothing to store */
 			RETURN_FALSE;
 		}
-		/* return true on success */
-		if (apc_cache_store(apc_user_cache, Z_STR_P(key), val, (uint32_t) ttl, exclusive)) {
-			RETURN_TRUE;
-		}
+
+		RETURN_BOOL(apc_cache_store(apc_user_cache, Z_STR_P(key), val, (uint32_t) ttl, exclusive));
 	} else {
 		apc_warning("apc_store expects key parameter to be a string or an array of key/value pairs.");
+		RETURN_FALSE;
 	}
-
-	/* default */
-	RETURN_FALSE;
 }
 /* }}} */
 
@@ -491,16 +499,17 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 	returns true when apcu is usable in the current environment */
 PHP_FUNCTION(apcu_enabled) {
 	RETURN_BOOL(APCG(enabled));
-}  /* }}} */
+}
+/* }}} */
 
-/* {{{ proto int apc_store(mixed key, mixed var [, long ttl ])
+/* {{{ proto int apcu_store(mixed key, mixed var [, long ttl ])
  */
 PHP_FUNCTION(apcu_store) {
 	apc_store_helper(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
 }
 /* }}} */
 
-/* {{{ proto int apc_add(mixed key, mixed var [, long ttl ])
+/* {{{ proto int apcu_add(mixed key, mixed var [, long ttl ])
  */
 PHP_FUNCTION(apcu_add) {
 	apc_store_helper(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
@@ -526,15 +535,15 @@ static zend_bool php_inc_updater(apc_cache_t* cache, apc_cache_entry_t* entry, v
 	return 0;
 }
 
-/* {{{ proto long apc_inc(string key [, long step [, bool& success]])
+/* {{{ proto long apcu_inc(string key [, long step [, bool& success [, long ttl]]])
  */
 PHP_FUNCTION(apcu_inc) {
 	zend_string *key;
 	struct php_inc_updater_args args;
-	zend_long step = 1;
+	zend_long step = 1, ttl = 0;
 	zval *success = NULL;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lz", &key, &step, &success) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lzl", &key, &step, &success, &ttl) == FAILURE) {
 		return;
 	}
 
@@ -545,7 +554,7 @@ PHP_FUNCTION(apcu_inc) {
 
 	ZVAL_LONG(&args.step, step);
 
-	if (php_apc_update(key, php_inc_updater, &args)) {
+	if (php_apc_update(key, php_inc_updater, &args, 1, ttl)) {
 		if (success) {
 			ZVAL_TRUE(success);
 		}
@@ -560,15 +569,15 @@ PHP_FUNCTION(apcu_inc) {
 }
 /* }}} */
 
-/* {{{ proto long apc_dec(string key [, long step [, bool &success]])
+/* {{{ proto long apcu_dec(string key [, long step [, bool &success [, long ttl]]])
  */
 PHP_FUNCTION(apcu_dec) {
 	zend_string *key;
 	struct php_inc_updater_args args;
-	zend_long step = 1;
+	zend_long step = 1, ttl = 0;
 	zval *success = NULL;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lz", &key, &step, &success) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lzl", &key, &step, &success, &ttl) == FAILURE) {
 		return;
 	}
 
@@ -579,7 +588,7 @@ PHP_FUNCTION(apcu_dec) {
 
 	ZVAL_LONG(&args.step, 0 - step);
 
-	if (php_apc_update(key, php_inc_updater, &args)) {
+	if (php_apc_update(key, php_inc_updater, &args, 1, ttl)) {
 		if (success) {
 			ZVAL_TRUE(success);
 		}
@@ -596,23 +605,15 @@ PHP_FUNCTION(apcu_dec) {
 /* }}} */
 
 /* {{{ php_cas_updater */
-static zend_bool php_cas_updater(apc_cache_t* cache, apc_cache_entry_t* entry, void* data) {
-	zend_long* vals = ((zend_long*)data);
+static zend_bool php_cas_updater(apc_cache_t *cache, zend_long *entry, void *data) {
+	zend_long *vals = (zend_long *) data;
 	zend_long old = vals[0];
 	zend_long new = vals[1];
-
-	if (Z_TYPE(entry->val) == IS_LONG) {
-		if (Z_LVAL(entry->val) == old) {
-			Z_LVAL(entry->val) = new;
-			return 1;
-		}
-	}
-
-	return 0;
+	return ATOMIC_CAS(*entry, old, new);
 }
 /* }}} */
 
-/* {{{ proto int apc_cas(string key, int old, int new)
+/* {{{ proto int apcu_cas(string key, int old, int new)
  */
 PHP_FUNCTION(apcu_cas) {
 	zend_string *key;
@@ -622,25 +623,22 @@ PHP_FUNCTION(apcu_cas) {
 		return;
 	}
 
-	if (php_apc_update(key, php_cas_updater, &vals)) {
-		RETURN_TRUE;
+	if (APCG(serializer_name)) {
+		/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
+		apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 	}
 
-	RETURN_FALSE;
+	RETURN_BOOL(apc_cache_atomic_update_long(apc_user_cache, key, php_cas_updater, &vals, 0, 0));
 }
 /* }}} */
 
-/* {{{ proto mixed apc_fetch(mixed key[, bool &success])
+/* {{{ proto mixed apcu_fetch(mixed key[, bool &success])
  */
 PHP_FUNCTION(apcu_fetch) {
 	zval *key;
 	zval *success = NULL;
 	time_t t;
 
-	if (!APCG(enabled)) {
-		RETURN_FALSE;
-	}
-
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z|z", &key, &success) == FAILURE) {
 		return;
 	}
@@ -657,57 +655,46 @@ PHP_FUNCTION(apcu_fetch) {
 		convert_to_string(key);
 	}
 
-	if (Z_TYPE_P(key) == IS_ARRAY || (Z_TYPE_P(key) == IS_STRING && Z_STRLEN_P(key) > 0)) {
-		if (Z_TYPE_P(key) == IS_STRING) {
-			if (apc_cache_fetch(apc_user_cache, Z_STR_P(key), t, &return_value)) {
-				if (success) {
-					ZVAL_TRUE(success);
-				}
-			} else { RETVAL_FALSE; }
-		} else if (Z_TYPE_P(key) == IS_ARRAY) {
-			zval *hentry;
-			zval result;
-
-			array_init(&result);
-			ZEND_HASH_FOREACH_VAL(Z_ARRVAL_P(key), hentry) {
-				ZVAL_DEREF(hentry);
-				if (Z_TYPE_P(hentry) == IS_STRING) {
-					zval result_entry,
-						*iresult = &result_entry;
-					ZVAL_UNDEF(iresult);
-
-					if (apc_cache_fetch(apc_user_cache, Z_STR_P(hentry), t, &iresult)) {
-						zend_symtable_update(Z_ARRVAL(result), Z_STR_P(hentry), &result_entry);
-					}
-				} else {
-					apc_warning("apc_fetch() expects a string or array of strings.");
-				}
-			} ZEND_HASH_FOREACH_END();
-
-			RETVAL_ZVAL(&result, 0, 1);
-
+	if (Z_TYPE_P(key) == IS_STRING) {
+		if (apc_cache_fetch(apc_user_cache, Z_STR_P(key), t, return_value)) {
 			if (success) {
 				ZVAL_TRUE(success);
 			}
+		} else { RETVAL_FALSE; }
+	} else if (Z_TYPE_P(key) == IS_ARRAY) {
+		zval *hentry;
+
+		array_init(return_value);
+		ZEND_HASH_FOREACH_VAL(Z_ARRVAL_P(key), hentry) {
+			ZVAL_DEREF(hentry);
+			if (Z_TYPE_P(hentry) == IS_STRING) {
+				zval result_entry;
+				ZVAL_UNDEF(&result_entry);
+
+				if (apc_cache_fetch(apc_user_cache, Z_STR_P(hentry), t, &result_entry)) {
+					zend_hash_update(Z_ARRVAL_P(return_value), Z_STR_P(hentry), &result_entry);
+				}
+			} else {
+				apc_warning("apc_fetch() expects a string or array of strings.");
+			}
+		} ZEND_HASH_FOREACH_END();
+
+		if (success) {
+			ZVAL_TRUE(success);
 		}
 	} else {
 		apc_warning("apc_fetch() expects a string or array of strings.");
 		RETURN_FALSE;
 	}
-	return;
 }
 /* }}} */
 
-/* {{{ proto mixed apc_exists(mixed key)
+/* {{{ proto mixed apcu_exists(mixed key)
  */
 PHP_FUNCTION(apcu_exists) {
 	zval *key;
 	time_t t;
 
-	if (!APCG(enabled)) {
-		RETURN_FALSE;
-	}
-
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z", &key) == FAILURE) {
 		return;
 	}
@@ -719,62 +706,42 @@ PHP_FUNCTION(apcu_exists) {
 	}
 
 	if (Z_TYPE_P(key) == IS_STRING) {
-		if (Z_STRLEN_P(key)) {
-			if (apc_cache_exists(apc_user_cache, Z_STR_P(key), t)) {
-				RETURN_TRUE;
-			} else {
-				RETURN_FALSE;
-			}
-		}
+		RETURN_BOOL(apc_cache_exists(apc_user_cache, Z_STR_P(key), t));
 	} else if (Z_TYPE_P(key) == IS_ARRAY) {
 		zval *hentry;
+		zval true_zv;
+		ZVAL_TRUE(&true_zv);
 
 		array_init(return_value);
 		ZEND_HASH_FOREACH_VAL(Z_ARRVAL_P(key), hentry) {
 			ZVAL_DEREF(hentry);
 			if (Z_TYPE_P(hentry) == IS_STRING) {
 				if (apc_cache_exists(apc_user_cache, Z_STR_P(hentry), t)) {
-					  add_assoc_bool(return_value, Z_STRVAL_P(hentry), 1);
-				   }
+					  zend_hash_add_new(Z_ARRVAL_P(return_value), Z_STR_P(hentry), &true_zv);
+				}
 			} else {
 				apc_warning(
 					"apc_exists() expects a string or array of strings.");
 			}
 		} ZEND_HASH_FOREACH_END();
-
-		return;
 	} else {
 		apc_warning("apc_exists() expects a string or array of strings.");
+		RETURN_FALSE;
 	}
-
-	RETURN_FALSE;
 }
 /* }}} */
 
-/* {{{ proto mixed apc_delete(mixed keys)
+/* {{{ proto mixed apcu_delete(mixed keys)
  */
 PHP_FUNCTION(apcu_delete) {
 	zval *keys;
 
-	if (!APCG(enabled)) {
-		RETURN_FALSE;
-	}
-
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z", &keys) == FAILURE) {
 		return;
 	}
 
 	if (Z_TYPE_P(keys) == IS_STRING) {
-		if (!Z_STRLEN_P(keys)) {
-			RETURN_FALSE;
-		}
-
-		if (apc_cache_delete(apc_user_cache, Z_STR_P(keys))) {
-			RETURN_TRUE;
-		} else {
-			RETURN_FALSE;
-		}
-
+		RETURN_BOOL(apc_cache_delete(apc_user_cache, Z_STR_P(keys)));
 	} else if (Z_TYPE_P(keys) == IS_ARRAY) {
 		zval *hentry;
 
@@ -782,7 +749,7 @@ PHP_FUNCTION(apcu_delete) {
 		ZEND_HASH_FOREACH_VAL(Z_ARRVAL_P(keys), hentry) {
 			ZVAL_DEREF(hentry);
 			if (Z_TYPE_P(hentry) != IS_STRING) {
-				apc_warning("apc_delete() expects a string, array of strings, or APCIterator instance.");
+				apc_warning("apc_delete() expects a string, array of strings, or APCIterator instance");
 				add_next_index_zval(return_value, hentry);
 				Z_TRY_ADDREF_P(hentry);
 			} else if (apc_cache_delete(apc_user_cache, Z_STR_P(hentry)) != 1) {
@@ -791,14 +758,10 @@ PHP_FUNCTION(apcu_delete) {
 			}
 		} ZEND_HASH_FOREACH_END();
 	} else if (Z_TYPE_P(keys) == IS_OBJECT) {
-
-		if (apc_iterator_delete(keys)) {
-			RETURN_TRUE;
-		} else {
-			RETURN_FALSE;
-		}
+		RETURN_BOOL(apc_iterator_delete(keys) != 0);
 	} else {
-		apc_warning("apc_delete() expects a string, array of strings, or APCIterator instance.");
+		apc_warning("apc_delete() expects a string, array of strings, or APCIterator instance");
+		RETURN_FALSE;
 	}
 }
 
@@ -817,6 +780,28 @@ PHP_FUNCTION(apcu_entry) {
 }
 /* }}} */
 
+#ifdef APC_DEBUG
+/* This function is used to test TTL behavior without having to perform sleeps. */
+PHP_FUNCTION(apcu_inc_request_time) {
+	zend_long by = 1;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &by) != SUCCESS) {
+		return;
+	}
+
+	if (!APCG(use_request_time)) {
+		php_error_docref(NULL, E_WARNING,
+			"Trying to increment request time while use_request_time is disabled");
+		return;
+	}
+
+	/* Ensure APCG(request_time) is primed */
+	(void) apc_time();
+
+	APCG(request_time) += by;
+}
+#endif
+
 /* {{{ apcu_functions[] */
 zend_function_entry apcu_functions[] = {
 	PHP_FE(apcu_cache_info,         arginfo_apcu_cache_info)
@@ -833,6 +818,9 @@ zend_function_entry apcu_functions[] = {
 	PHP_FE(apcu_cas,                arginfo_apcu_cas)
 	PHP_FE(apcu_exists,             arginfo_apcu_exists)
 	PHP_FE(apcu_entry,				arginfo_apcu_entry)
+#ifdef APC_DEBUG
+	PHP_FE(apcu_inc_request_time,   NULL)
+#endif
 	PHP_FE_END
 };
 /* }}} */
